#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include <GL/glew.h>
#include <GL/glu.h>

#include <SDL.h>
#include <SDL_opengl.h>
#include <SDL_image.h>

#include "utils_graphics.h"


#define MAX_UNITS           6.0f
#define SCREEN_WIDTH        640
#define SCREEN_HEIGHT       480
#define MAX_STACK_ITEMS     100
#define STARS_COUNT         50

#define POSITION            0
#define TEX_COORD           1

#define DEG_TO_RAD(degrees) (degrees) * M_PI / 180




struct _matrix_stack {
    int items_count;
    float items[MAX_STACK_ITEMS][16];
} matrix_stack;

typedef struct _star {
    float angle;
    float dist;
    float rotation_speed;

    float r;
    float g;
    float b;
} star;

SDL_Window* gWindow = NULL;
SDL_GLContext gContext;
SDL_Surface *image2d;

unsigned int last_time;
float tilt = 90.0f;
float spin = 0.0f;
float effectiveFPMS = 60.0f / 1000.0f;
star *stars = 0;

GLuint vertex_shader;
GLuint fragment_shader;
GLuint shader_program;
GLuint a_position_loc;
GLuint a_texture_coords_loc;
GLuint u_mvpMatrix_loc;

GLuint vertex_array;
GLuint vertex_buffer;
GLuint texture_coords_buffer;

float projection_matrix[16];
float modelview_matrix[16];
float mvp_matrix[16];
float aux_matrix1[16];
float aux_matrix2[16];

GLfloat left = -MAX_UNITS;
GLfloat right = MAX_UNITS;
GLfloat bottom = -MAX_UNITS;
GLfloat top = MAX_UNITS;
GLfloat near = MAX_UNITS + 1.0f;
GLfloat far = MAX_UNITS + 5.0f;

int texture;

GLfloat vertices[4*3] = {
    -1.0, -1.0,  0.0,
     1.0, -1.0,  0.0,
    -1.0,  1.0,  0.0,
     1.0,  1.0,  0.0
};

GLfloat tex_coords[4*2] = {
    0.0, 0.0,
    1.0, 0.0,
    0.0, 1.0,
    1.0, 1.0
};




int init();
int initGL();
int check_gl_error();
void cleanup();
int set_shaders();
int init_buffers();
void update();
void render();

void push_matrix(float matrix[16]);
void pop_matrix(float matrix[16]);

char *text_file_read(char *fname);
void print_shader_info_log(GLuint obj);
void print_program_info_log(GLuint obj);



void push_matrix(float matrix[16]) {
    if (matrix_stack.items_count + 1 <= MAX_STACK_ITEMS) {
        memcpy(matrix_stack.items[matrix_stack.items_count], matrix, 16 * sizeof(float));
        matrix_stack.items_count++;
    }
}

void pop_matrix(float matrix[16]) {
    if (matrix_stack.items_count - 1 >= 0) {
        matrix_stack.items_count--;
        memcpy(matrix, matrix_stack.items[matrix_stack.items_count], 16 * sizeof(float));
    }
}

int init() {
    // Initialize SDL
    if (SDL_Init(SDL_INIT_VIDEO) < 0) {
        printf("Błąd podczas inicjalizacji silnika SDL2: %s\n", SDL_GetError());
        return 0;
    }

    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 1);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);

    gWindow = SDL_CreateWindow("Wirujące gwiazdy", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED,
                                SCREEN_WIDTH, SCREEN_HEIGHT, SDL_WINDOW_OPENGL | SDL_WINDOW_SHOWN);
    if (gWindow == NULL) {
        printf("Błąd podczas tworzenia okna: %s\n", SDL_GetError());
        return 0;
    }

    gContext = SDL_GL_CreateContext(gWindow);
    if (gContext == NULL) {
        printf("Błąd podczas tworzenia kontekstu OpenGL: %s\n", SDL_GetError());
        return 0;
    }

    if (SDL_GL_SetSwapInterval(1) < 0) {
        printf( "Błąd podczas ustawiania VSync: %s\n", SDL_GetError());
    }

    int imgFlags = IMG_INIT_PNG;
    if (!(IMG_Init(imgFlags) & imgFlags)) {
        printf("Błąd podczas inicjalizacji biblioteki SDL_image: %s\n", IMG_GetError());
        return 0;
    }

    char *path = "star.png";
    image2d = IMG_Load(path);
    if (image2d == NULL) {
        printf("Błąd podczas ładowania pliku tekstury %s: %s\n", path, SDL_GetError());
        return 0;
    }

    if (!initGL()) {
        printf("Błąd podczas inicjalizacji OpenGL!\n");
        return 0;
    }

    stars = malloc(STARS_COUNT * sizeof(star));
    if (!stars) {
        return 0;
    }
    memset(stars, 0, STARS_COUNT * sizeof(star));
    for (int i = 0; i < STARS_COUNT; i++) {
        stars[i].angle = 0.0f;
        stars[i].dist = ((float)i / (float)STARS_COUNT) * 5.0f;
        stars[i].rotation_speed = (float)i / (float)STARS_COUNT;
        stars[i].r = 1.0f * (float)rand() / (float)RAND_MAX;
        stars[i].g = 1.0f * (float)rand() / (float)RAND_MAX;
        stars[i].b = 1.0f * (float)rand() / (float)RAND_MAX;
    }
    last_time = SDL_GetTicks();
    return 1;
}

int initGL() {
    GLenum err = GL_NO_ERROR;
    GLfloat fovy = 45.0;
    GLfloat aspect = 1.0f;
    unsigned int width = SCREEN_WIDTH;
    unsigned int height = SCREEN_HEIGHT;

    glewExperimental = GL_TRUE;
    err = glewInit();
    while (check_gl_error() == GL_INVALID_ENUM) {};
    if (err != GLEW_OK) {
        printf("glewInit failed: %s\n", glewGetErrorString(err));
        return 0;
    }

    glClearColor(0.f, 0.f, 0.f, 1.f);
    glEnable(GL_DEPTH_TEST);

    printf("surface->format->BytesPerPixel: %d\n", image2d->format->BytesPerPixel);
    glGenTextures(1, &texture);
    glBindTexture(GL_TEXTURE_2D, texture);
    //glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
    //gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, image2d->w, image2d->h, GL_RGBA, GL_UNSIGNED_BYTE, image2d->pixels);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, image2d->w, image2d->h, 0, GL_RGBA, GL_UNSIGNED_BYTE, image2d->pixels);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    if (image2d) {
        SDL_FreeSurface(image2d);
        image2d = NULL;
    }
    glBindTexture(GL_TEXTURE_2D, 0);

    matrix_stack.items_count = 0;

    glViewport(0, 0, width, height);

    if (height > 0) {
        aspect = (float)width / (float)height;
    }

    get_perspective_matrix(projection_matrix, fovy, aspect, near, far);

    if (!set_shaders()) {
        cleanup();
        return 0;
    }
    if (!init_buffers()) {
        cleanup();
        return 0;
    }

    return 1;
}


int set_shaders() {
    char *shader_source = NULL;
    GLint status;

    shader_source = text_file_read("particles_01.vshader");
    if (!shader_source) {
        printf("Nie można odczytać shadera wierzchołków\n");
        return 0;
    }
    //printf("%s\n---\n", shader_source);
    vertex_shader = glCreateShader(GL_VERTEX_SHADER);
    //printf("vertex_shader: %d\n", vertex_shader);
    glShaderSource(vertex_shader,  1, (const char* const*) &shader_source, NULL);
    free(shader_source);
    glCompileShader(vertex_shader);
    glGetShaderiv(vertex_shader, GL_COMPILE_STATUS, &status);
    if (status != GL_TRUE) {
        printf("Niepoprawna kompilacja shadera wierzchołków\n");
        print_shader_info_log(vertex_shader);
        return 0;
    }

    shader_source = text_file_read("particles_01.fshader");
    if (!shader_source) {
        printf("Nie można odczytać shadera fragmentów\n");
        return 0;
    }
    //printf("%s\n---\n", shader_source);
    fragment_shader = glCreateShader(GL_FRAGMENT_SHADER);
    //printf("fragment_shader: %d\n", fragment_shader);
    glShaderSource(fragment_shader,  1, (const char* const*) &shader_source, NULL);
    free(shader_source);
    glCompileShader(fragment_shader);
    glGetShaderiv(fragment_shader, GL_COMPILE_STATUS, &status);
    if (status != GL_TRUE) {
        printf("Niepoprawna kompilacja shadera fragmentów\n");
        print_shader_info_log(vertex_shader);
        return 0;
    }

    shader_program = glCreateProgram();
    glAttachShader(shader_program, vertex_shader);
    glAttachShader(shader_program, fragment_shader);
    glLinkProgram(shader_program);
    glGetProgramiv(shader_program, GL_LINK_STATUS, &status);
    if (status == GL_FALSE) {
        printf("Niepoprawna konsolidacja programu\n");
        print_program_info_log(shader_program);
        return 0;
    }
    glValidateProgram(shader_program);
    glGetProgramiv(shader_program, GL_VALIDATE_STATUS, &status);
    if (status == GL_FALSE) {
        printf("Niepoprawna walidacja programu\n");
        print_program_info_log(shader_program);
        return 0;
    }

    glUseProgram(shader_program);

    return 1;
}


int init_buffers() {
    glGenVertexArrays(1, &vertex_array);
    glBindVertexArray(vertex_array);

    glGenBuffers(1, &vertex_buffer);
    glBindBuffer(GL_ARRAY_BUFFER, vertex_buffer);
    glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW);
    a_position_loc = glGetAttribLocation(shader_program, "in_position");
    glVertexAttribPointer(a_position_loc, 3, GL_FLOAT, GL_FALSE, 0, NULL);
    glEnableVertexAttribArray(a_position_loc);

    glGenBuffers(1, &texture_coords_buffer);
    glBindBuffer(GL_ARRAY_BUFFER, texture_coords_buffer);
    glBufferData(GL_ARRAY_BUFFER, sizeof(tex_coords), tex_coords, GL_STATIC_DRAW);
    a_texture_coords_loc = glGetAttribLocation(shader_program, "in_tex_coord");
    glVertexAttribPointer(a_texture_coords_loc, 2, GL_FLOAT, GL_FALSE, 0, NULL);
    glEnableVertexAttribArray(a_texture_coords_loc);

    glBindVertexArray(0);

    //printf("%d, %d\n", sizeof(vertices), sizeof(tex_coords));

    return 1;
}

void update() {
    int i;
    unsigned int time_now = SDL_GetTicks();

    if (last_time != 0) {
        unsigned long elapsed = time_now - last_time;
        for (i = 0; i < STARS_COUNT; i++) {
            stars[i].angle += stars[i].rotation_speed * effectiveFPMS * elapsed;
            stars[i].dist -= 0.01 * effectiveFPMS * elapsed;
            if (stars[i].dist < 0.0) {
                stars[i].dist += 5.0;
                stars[i].r = 1.0f * (float)rand() / (float)RAND_MAX;
                stars[i].g = 1.0f * (float)rand() / (float)RAND_MAX;
                stars[i].b = 1.0f * (float)rand() / (float)RAND_MAX;
            }
        }
    }
    last_time = time_now;
}

void render() {
    int i;

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glDepthMask(GL_FALSE);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);

    get_identity_matrix(modelview_matrix);


    get_translation_matrix4x4(aux_matrix1, 0.0f, 0.0f, -(near+far)/2.0f);
    mult_matrices(aux_matrix2, modelview_matrix, aux_matrix1, 4);
    memcpy(modelview_matrix, aux_matrix2, 16 * sizeof(float));

    get_rotation_x_matrix4x4(aux_matrix1, DEG_TO_RAD(tilt));
    mult_matrices(aux_matrix2, modelview_matrix, aux_matrix1, 4);
    memcpy(modelview_matrix, aux_matrix2, 16 * sizeof(float));

    for (i = 0; i < STARS_COUNT; i++) {
        //dump_matrix4x4(modelview_matrix);
        push_matrix(modelview_matrix);

        get_rotation_y_matrix4x4(aux_matrix1, DEG_TO_RAD(stars[i].angle));
        mult_matrices(aux_matrix2, modelview_matrix, aux_matrix1, 4);

        get_translation_matrix4x4(aux_matrix1, stars[i].dist, 0.0f, 0.0f);
        mult_matrices(modelview_matrix, aux_matrix2, aux_matrix1, 4);

        get_rotation_y_matrix4x4(aux_matrix1, DEG_TO_RAD(-stars[i].angle));
        mult_matrices(aux_matrix2, modelview_matrix, aux_matrix1, 4);

        get_rotation_x_matrix4x4(aux_matrix1, DEG_TO_RAD(-tilt));
        mult_matrices(aux_matrix2, modelview_matrix, aux_matrix1, 4);

        get_rotation_z_matrix4x4(aux_matrix1, DEG_TO_RAD(spin));
        mult_matrices(modelview_matrix, aux_matrix2, aux_matrix1, 4);


        mult_matrices(mvp_matrix, projection_matrix, modelview_matrix, 4);
        transpose_matrix(aux_matrix1, mvp_matrix, 4);
        memcpy(mvp_matrix, aux_matrix1, 16 * sizeof(float));

        glBindVertexArray(vertex_array);
        glUseProgram(shader_program);
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, texture);
        glUniform1i(glGetUniformLocation(shader_program, "tex"), 0);

        glUniformMatrix4fv(glGetUniformLocation(shader_program, "u_mvp_matrix"), 1, GL_FALSE, mvp_matrix);
        glUniform3f(glGetUniformLocation(shader_program, "u_color"), stars[i].r, stars[i].g, stars[i].b);

        glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

        glUseProgram(0);
        glBindVertexArray(0);

        pop_matrix(modelview_matrix);
        //printf("%.2f, %.2f, %.2f, %.2f, %.2f, %.2f\n", stars[i].angle, stars[i].dist, stars[i].rotation_speed, stars[i].r, stars[i].g, stars[i].b);
        //dump_matrix4x4(modelview_matrix);
        spin += 0.1f;
    }

    glDisable(GL_BLEND);
    glDepthMask(GL_TRUE);
}

void cleanup() {
    if (stars) {
        free(stars);
    }
    if (vertex_buffer) {
        glDeleteBuffers(1, &vertex_buffer);
    }
    if (vertex_array) {
        glDeleteVertexArrays(1, &vertex_array);
    }

    if (shader_program && fragment_shader) {
        glDetachShader(shader_program, fragment_shader);
    }
    if (shader_program && vertex_shader) {
        glDetachShader(shader_program, vertex_shader);
    }
    if (fragment_shader) {
        glDeleteShader(fragment_shader);
    }
    if (vertex_shader) {
        glDeleteShader(vertex_shader);
    }
    if (shader_program) {
        glDeleteProgram(shader_program);
    }

    glDeleteTextures(1, &texture);
    if (image2d) {
        SDL_FreeSurface(image2d);
        image2d = NULL;
    }

    if (gContext) {
        SDL_GL_DeleteContext(gContext);
        gContext = NULL;
    }

    if (gWindow) {
        SDL_DestroyWindow(gWindow);
        gWindow = NULL;
    }

    IMG_Quit();
    SDL_Quit();
}



int main(int argc, char* args[]) {
    int quit;
    SDL_Event e;

    if (!init()) {
        printf("Błąd inicjalizacji!\n");
        cleanup();
        return 1;
    }

    srand(time(0));

    quit = 0;
    while (!quit) {
        while (SDL_PollEvent(&e) != 0) {
            if (e.type == SDL_QUIT) {
                quit = 1;
                break;
            }
        }

        update();
        render();

        SDL_GL_SwapWindow(gWindow);
    }

    cleanup();

    return 0;
}



char *text_file_read(char *fname) {
    FILE *fp;
    char *content = NULL;
    int count = 0;

    if (fname != NULL) {
        fp = fopen(fname, "rt");

        if (fp != NULL) {
            fseek(fp, 0, SEEK_END);
            count = ftell(fp);
            rewind(fp);

            if (count > 0) {
                content = (char*)malloc(sizeof(char) * (count+1));
                count = fread(content,sizeof(char),count,fp);
                content[count] = '\0';
            }
            fclose(fp);
        }
    }
    return content;
}


void print_shader_info_log(GLuint obj) {
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

    glGetShaderiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

    if (infologLength > 0) {
        infoLog = (char*)malloc(infologLength);
        glGetShaderInfoLog(obj, infologLength, &charsWritten, infoLog);
        printf("%s\n",infoLog);
        free(infoLog);
    }
}

void print_program_info_log(GLuint obj) {
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

    glGetProgramiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

    if (infologLength > 0) {
        infoLog = (char*)malloc(infologLength);
        glGetProgramInfoLog(obj, infologLength, &charsWritten, infoLog);
        printf("%s\n",infoLog);
        free(infoLog);
    }
}


int check_gl_error() {
    GLenum error;

    error = glGetError();
    switch (error) {
        case GL_CONTEXT_LOST:
            printf("GL_CONTEXT_LOST [%d]: %s\n", error, gluErrorString(error));
            return GL_CONTEXT_LOST;
        case GL_INVALID_ENUM:
            printf("GL_INVALID_ENUM [%d]: %s\n", error, gluErrorString(error));
            return GL_INVALID_ENUM;
        case GL_INVALID_VALUE:
            printf("GL_INVALID_VALUE [%d]: %s\n", error, gluErrorString(error));
            return GL_INVALID_VALUE;
        case GL_INVALID_OPERATION:
            printf("GL_INVALID_OPERATION [%d]: %s\n", error, gluErrorString(error));
            return 1;
        case GL_INVALID_FRAMEBUFFER_OPERATION:
            printf("GL_INVALID_FRAMEBUFFER_OPERATION [%d]: %s\n", error, gluErrorString(error));
            return GL_INVALID_FRAMEBUFFER_OPERATION;
        case GL_OUT_OF_MEMORY:
            printf("GL_OUT_OF_MEMORY [%d]: %s\n", error, gluErrorString(error));
            return 1;
        case GL_STACK_OVERFLOW:
            printf("GL_STACK_OVERFLOW [%d]: %s\n", error, gluErrorString(error));
            return GL_STACK_OVERFLOW;
        case GL_STACK_UNDERFLOW:
            printf("GL_STACK_UNDERFLOW [%d]: %s\n", error, gluErrorString(error));
            return GL_STACK_UNDERFLOW;
        case GL_NO_ERROR:
            printf("GL_NO_ERROR [%d]\n", error);
            break;
    }

    return GL_NO_ERROR;
}
